package exec

import (
	"errors"
	"fmt"
	"os/exec"
	"path/filepath"
	"strings"
)

// Checks guestfish is available.
func CheckGuestfish() error {
	output, err := exec.Command(conf.Tools.Guestfish, "--version").Output()
	if err != nil {
		return errors.New(conf.Tools.Guestfish + " is required but not found. On Ubuntu/Debian, it can be installed with \"sudo apt-get install guestfish\".")
	}
	out := string(output)
	lines := strings.Split(out, "\n")
	fields := strings.Split(lines[0], " ")
	if len(fields) < 2 {
		return errors.New("Failed extracting " + conf.Tools.Guestfish + " version number!")
	}
	return nil
}

// Uploads and unarchives a local tar.gz archive into a directory (vmDir) inside the VM's filesystem.
func DearchiveToVM(vmDiskFile, localTarGzFile, vmDir string) error {
	cmd := exec.Command(conf.Tools.Guestfish, "--rw", "-i", "tar-in", "-a", vmDiskFile, localTarGzFile, vmDir, "compress:gzip")
	stdoutStderr, err := cmd.CombinedOutput()
	if err != nil {
		output := fmt.Sprintf("[%s]", stdoutStderr)
		msg := "Failed dearchiving to \"" + vmDir + "\" in VM"
		log.Error(msg + ": " + output)
		return errors.New(msg)
	}
	return nil
}

// Uploads a local file into a directory (vmDir) inside the VM's filesystem.
func CopyToVM(vmDiskFile, localFile, vmDir string) error {
	filename := filepath.Base(localFile)
	cmd := exec.Command(conf.Tools.Guestfish, "--rw", "-i", "upload", "-a", vmDiskFile, localFile, filepath.Join(vmDir, filename))
	stdoutStderr, err := cmd.CombinedOutput()
	if err != nil {
		output := fmt.Sprintf("[%s]", stdoutStderr)
		msg := "Failed copying file to \"" + vmDir + "\" in VM"
		log.Error(msg + ": " + output)
		return errors.New(msg)
	}
	return nil
}

// Recursively downloads a directory from the VM's filesystem (vmDir) into a local tar.gz archive.
func ArchiveFromVM(vmDiskFile, vmDir, localTarGzFile string) error {
	cmd := exec.Command(conf.Tools.Guestfish, "--ro", "-i", "tar-out", "-a", vmDiskFile, vmDir, localTarGzFile, "compress:gzip")
	stdoutStderr, err := cmd.CombinedOutput()
	if err != nil {
		output := fmt.Sprintf("[%s]", stdoutStderr)
		msg := "Failed archiving \"" + vmDir + "\" in VM"
		log.Error(msg + ": " + output)
		return errors.New(msg)
	}
	return nil
}

// Delete a file from a VM's filesystem.
// filePath is the full path in the VM to the file to delete.
func DeleteFromVM(vmDiskFile, filePath string) error {
	cmd := exec.Command(conf.Tools.Guestfish, "--rw", "-a", vmDiskFile, "-i", "rm", filePath)
	stdoutStderr, err := cmd.CombinedOutput()
	if err != nil {
		output := fmt.Sprintf("[%s]", stdoutStderr)
		msg := "Failed deleting \"" + filePath + "\" in VM"
		log.Error(msg + ": " + output)
		return errors.New(msg)
	}
	return nil
}
