package router

import (
	"io"
	"net/http"
	"os"
	"path/filepath"

	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/server/users"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/server/vms"

	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/common/caps"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/common/params"

	"github.com/google/uuid"
	"github.com/labstack/echo/v4"
)

type (
	RouterTemplates struct {
		users *users.Users
		tpl   *vms.Templates
		vms   *vms.VMs
	}
)

func NewRouterTemplates() *RouterTemplates {
	return &RouterTemplates{users: users.GetUsersInstance(), tpl: vms.GetTemplatesInstance(), vms: vms.GetVMsInstance()}
}

// Returns templates that can be listed.
// Requires either capability:
// CAP_TPL_LIST_ANY: any templates can be listed.
// CAP_TPL_LIST: only templates owned by the user as well as public templates can be listed.
// curl --cacert ca.pem -X GET https://localhost:1077/templates -H "Authorization: Bearer <AccessToken>"
func (r *RouterTemplates) GetTemplates(c echo.Context) error {
	// Retrieves logged user from context.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	// If the logged user has CAP_TPL_LIST_ANY, lists all templates.
	if user.HasCapability(caps.USERCAP_TPL_LIST_ANY) {
		// Returns all templates
		return c.JSONPretty(http.StatusOK, r.tpl.GetNetworkSerializedTemplates(func(template *vms.Template) bool { return true }), "    ")
	} else if user.HasCapability(caps.USERCAP_TPL_LIST) {
		// Returns templates owned by the logged user and public templates.
		return c.JSONPretty(http.StatusOK,
			r.tpl.GetNetworkSerializedTemplates(func(template *vms.Template) bool {
				return template.GetOwner() == user.Email || template.IsPublic()
			}), "    ")
	} else {
		return echo.NewHTTPError(http.StatusForbidden, msgInsufficientCaps)
	}
}

// Returns a template that can be listed based on its UUID.
// Requires either capability:
// CAP_TPL_LIST_ANY: any templates can be listed.
// CAP_TPL_LIST: only templates owned by the user as well as public templates can be listed.
// curl --cacert ca.pem -X GET https://localhost:1077/templates/62ae8791-c108-4235-a7d6-074e9b6a9017 -H "Authorization: Bearer <AccessToken>"
func (r *RouterTemplates) GetTemplate(c echo.Context) error {
	// Retrieves logged user from context.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	// Retrieves the template based on its UUID.
	id, err := uuid.Parse(c.Param("id"))
	if err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}
	tpl, err := r.tpl.GetTemplate(id)
	if err != nil {
		return echo.NewHTTPError(http.StatusNotFound, err.Error())
	}

	// If user has CAP_TPL_LIST_ANY, returns the template.
	if user.HasCapability(caps.USERCAP_TPL_LIST_ANY) {
		return c.JSONPretty(http.StatusOK, tpl.SerializeToNetwork(), "    ")
	} else if user.HasCapability(caps.USERCAP_TPL_LIST) {
		// Returns template if owned by the user or template is public.
		if tpl.GetOwner() == user.Email || tpl.IsPublic() {
			return c.JSONPretty(http.StatusOK, tpl.SerializeToNetwork(), "    ")
		}
	}

	return echo.NewHTTPError(http.StatusForbidden, msgInsufficientCaps)
}

// Creates a template from a VM.
// Requires CAP_TPL_CREATE.
// curl --cacert ca.pem -X POST https://localhost:1077/templates/vm -H 'Content-Type: application/json' -d '{"vmID":"e41f3556-ca24-4658-bd79-8c85bd6bff59","name":"Xubuntu_22.04","access":"public"}' -H "Authorization: Bearer <AccessToken>"
func (r *RouterTemplates) CreateTemplateFromVM(c echo.Context) error {
	// Retrieves logged user from context.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	if !user.HasCapability(caps.USERCAP_TPL_CREATE) {
		return echo.NewHTTPError(http.StatusForbidden, msgInsufficientCaps)
	}

	// Deserializes and validates client's parameters.
	p := new(params.TplCreate)
	if err := decodeJson(c, &p); err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	// Retrieves the VM from its vmID and checks the logged user can access it.
	vm, err := r.vms.GetVM(p.VMID)
	if err != nil {
		return echo.NewHTTPError(http.StatusNotFound, err.Error())
	}

	// Creates a new template from the client's parameters.
	template, err := vms.NewTemplateFromVM(p.Name, user.Email, p.Access, vm)
	if err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	// Adds the template.
	if err = r.tpl.AddTemplate(template); err != nil {
		return echo.NewHTTPError(http.StatusInternalServerError, err.Error())
	}

	return c.JSONPretty(http.StatusCreated, template.SerializeToNetwork(), "    ")
}

// Creates a template from an uploaded .qcow file.
// Requires CAP_TPL_CREATE.
// curl --cacert ca.pem -X POST https://localhost:1077/templates/qcow -H 'Content-Type: multipart/form-data' -F name="Xubuntu_22.04" -F access=public -F qcow=@disk.qcow -H "Authorization: Bearer <AccessToken>"
//
// IMPORTANT:
// There is a bug in Echo's upload when using HTTPS (TLS):
// If the uploaded file is >= 34MB, echo leaves a multipart-xxx file in /tmp
// If the uploaded file is <= 33MB, the multipart-xxx file is deleted as it should.
// There is no such problem when using HTTP.
// The upload uses exactly the code recipe desribed here:
// https://echo.labstack.com/guide/#form-multipartform-data
func (r *RouterTemplates) CreateTemplateFromQCOW(c echo.Context) error {
	// Retrieves logged user from context.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	if !user.HasCapability(caps.USERCAP_TPL_CREATE) {
		return echo.NewHTTPError(http.StatusForbidden, msgInsufficientCaps)
	}

	name := c.FormValue("name")
	access := c.FormValue("access")
	if err = vms.ValidateTemplateAccess(access); err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	qcowFile, err := c.FormFile("qcow")
	if err != nil {
		return echo.NewHTTPError(http.StatusInternalServerError, err.Error())
	}
	src, err := qcowFile.Open()
	if err != nil {
		return echo.NewHTTPError(http.StatusInternalServerError, err.Error())
	}
	defer src.Close()

	uuid, err := uuid.NewRandom()
	if err != nil {
		log.Error("Failed creating random UUID: " + err.Error())
		return echo.NewHTTPError(http.StatusInternalServerError, err.Error())
	}
	uploadedFile := filepath.Join(conf.Core.TmpDir, "upload_"+uuid.String()+".qcow")
	dst, err := os.Create(uploadedFile)
	if err != nil {
		return echo.NewHTTPError(http.StatusInternalServerError, err.Error())
	}
	defer dst.Close()
	defer os.Remove(uploadedFile)

	if _, err = io.Copy(dst, src); err != nil {
		return echo.NewHTTPError(http.StatusInternalServerError, err.Error())
	}

	// Creates a new template
	template, err := vms.NewTemplateFromQCOW(name, user.Email, access, uploadedFile)
	if err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	// Adds the template.
	if err = r.tpl.AddTemplate(template); err != nil {
		return echo.NewHTTPError(http.StatusInternalServerError, err.Error())
	}

	return c.JSONPretty(http.StatusCreated, template.SerializeToNetwork(), "    ")
}

// Deletes a template based on its ID.
// Requires either of:
// CAP_TPL_DESTROY_ANY: any template can be deleted.
// CAP_TPL_DESTROY: only a template owned by the user can be deleted.
// REMARK: a template can only be deleted if no VM references it!
// curl --cacert ca.pem -X DELETE https://localhost:1077/templates/4913a2bb-edfe-4dfe-af53-38197a44523b -H "Authorization: Bearer <AccessToken>"
func (r *RouterTemplates) DeleteTemplate(c echo.Context) error {
	// Retrieves logged user from context.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	// Retrieves template ID.
	tplID, err := uuid.Parse(c.Param("id"))
	if err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	if user.HasCapability(caps.USERCAP_TPL_DESTROY_ANY) {
		if err := r.tpl.DeleteTemplate(tplID, r.vms); err != nil {
			return echo.NewHTTPError(http.StatusNotFound, err.Error())
		}
		return c.NoContent(http.StatusOK)
	} else if user.HasCapability(caps.USERCAP_TPL_DESTROY) {
		template, err := r.tpl.GetTemplate(tplID)
		if err != nil {
			return echo.NewHTTPError(http.StatusNotFound, err.Error())
		}

		if template.GetOwner() == user.Email {
			if err := r.tpl.DeleteTemplate(tplID, r.vms); err != nil {
				return echo.NewHTTPError(http.StatusNotFound, err.Error())
			}
			return c.NoContent(http.StatusOK)
		}
	}

	return echo.NewHTTPError(http.StatusForbidden, msgInsufficientCaps)
}

// Edit a template based on its ID.
// Requires either of:
// CAP_TPL_EDIT_ANY: any template can be edited.
// CAP_TPL_EDIT: only a template owned by the user can be edited.
// curl --cacert ca.pem -X PUT https://localhost:1077/templates/4913a2bb-edfe-4dfe-af53-38197a44523b -H "Authorization: Bearer <AccessToken>"
func (r *RouterTemplates) EditTemplate(c echo.Context) error {
	// Retrieves logged user from context.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	// Retrieves template ID.
	tplID, err := uuid.Parse(c.Param("id"))
	if err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	editTemplate := func(c echo.Context, tplID uuid.UUID) error {
		// Deserializes and validates client's parameters.
		// Given these parameters are optional, we can't use a validator on them.
		// Validation is performed in templates.EditTemplate() instead.
		p := new(params.TplEdit)
		if err := decodeJson(c, &p); err != nil {
			return echo.NewHTTPError(http.StatusBadRequest, err.Error())
		}
		if err := r.tpl.EditTemplate(tplID, p.Name, p.Access); err != nil {
			return echo.NewHTTPError(http.StatusBadRequest, err.Error())
		}

		editedTpl, err := r.tpl.GetTemplate(tplID)
		if err != nil {
			return echo.NewHTTPError(http.StatusNotFound, err.Error())
		}
		return c.JSONPretty(http.StatusOK, editedTpl.SerializeToNetwork(), "    ")
	}

	if user.HasCapability(caps.USERCAP_TPL_EDIT_ANY) {
		return editTemplate(c, tplID)
	} else if user.HasCapability(caps.USERCAP_TPL_EDIT) {
		template, err := r.tpl.GetTemplate(tplID)
		if err != nil {
			return echo.NewHTTPError(http.StatusNotFound, err.Error())
		}
		if template.GetOwner() == user.Email {
			return editTemplate(c, tplID)
		}
	}

	return echo.NewHTTPError(http.StatusForbidden, msgInsufficientCaps)
}

// Exports a template's disk image.
// Requires either of:
// CAP_TPL_READFS_ANY: any template can have its disk exported.
// CAP_TPL_READFS: only a template owned by the user can have its disk exported.
// curl --cacert ca.pem -X GET https://localhost:1077/templates/4913a2bb-edfe-4dfe-af53-38197a44523b/disk -H "Authorization: Bearer <AccessToken>" --output disk.qcow
func (r *RouterTemplates) ExportDisk(c echo.Context) error {
	// Retrieves logged user from context.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	// Retrieves template ID.
	tplID, err := uuid.Parse(c.Param("id"))
	if err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	if user.HasCapability(caps.USERCAP_TPL_READFS_ANY) {
		template, err := r.tpl.GetTemplate(tplID)
		if err != nil {
			return echo.NewHTTPError(http.StatusNotFound, err.Error())
		}
		return c.File(template.GetTemplateDiskPath())

	} else if user.HasCapability(caps.USERCAP_TPL_READFS) {
		template, err := r.tpl.GetTemplate(tplID)
		if err != nil {
			return echo.NewHTTPError(http.StatusNotFound, err.Error())
		}
		if template.IsPublic() || template.GetOwner() == user.Email {
			return c.File(template.GetTemplateDiskPath())
		}
	}

	return echo.NewHTTPError(http.StatusForbidden, msgInsufficientCaps)
}
