package router

import (
	"net/http"

	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/server/users"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/server/utils"

	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/common/caps"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/common/params"

	"github.com/go-playground/validator/v10"
	"github.com/labstack/echo/v4"
)

type RouterUsers struct {
	users *users.Users
}

func NewRouterUsers() *RouterUsers {
	return &RouterUsers{users.GetUsersInstance()}
}

// Lists users.
// Requires CAP_USER_LIST.
// curl --cacert ca.pem -X GET http://localhost:1077/users -H "Authorization: Bearer <AccessToken>"
func (r *RouterUsers) GetUsers(c echo.Context) error {
	// Retrieves logged user from context.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	if !user.HasCapability(caps.USERCAP_USER_LIST) {
		return echo.NewHTTPError(http.StatusForbidden, msgInsufficientCaps)
	}

	list := []params.UserWithoutPwd{}
	for _, u := range r.users.GetUsers() {
		user := params.UserWithoutPwd{Email: u.Email, FirstName: u.FirstName, LastName: u.LastName, Caps: u.Caps}
		list = append(list, user)
	}

	return c.JSONPretty(http.StatusOK, list, "    ")
}

// Retrieves the logged (authententicated) user.
// curl --cacert ca.pem -X GET http://localhost:1077/users/me -H "Authorization: Bearer <AccessToken>"
func (r *RouterUsers) GetLoggedUser(c echo.Context) error {
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	userWithoutPwd := params.UserWithoutPwd{Email: user.Email, FirstName: user.FirstName, LastName: user.LastName, Caps: user.Caps}
	return c.JSONPretty(http.StatusOK, userWithoutPwd, "    ")
}

// Deletes a user by its ID.
// Requires CAP_USER_DESTROY.
// curl --cacert ca.pem -X DELETE http://localhost:1077/users/janedoes@nexus.org -H "Authorization: Bearer <AccessToken>"
func (r *RouterUsers) DeleteUserByEmail(c echo.Context) error {
	// Retrieves logged user from context.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	if !user.HasCapability(caps.USERCAP_USER_DESTROY) {
		return echo.NewHTTPError(http.StatusForbidden, msgInsufficientCaps)
	}

	userToDelete, err := r.users.GetUserByEmail(c.Param("email"))
	if err != nil {
		return echo.NewHTTPError(http.StatusNotFound, err.Error())
	}

	if err = r.users.DeleteUserByEmail(userToDelete.Email); err != nil {
		return echo.NewHTTPError(http.StatusNotFound, err.Error())
	}

	return c.NoContent(http.StatusOK)
}

// Creates a user.
// Requires CAP_USER_CREATE.
// curl --cacert ca.pem -X POST http://localhost:1077/users -H 'Content-Type: application/json' -d '{"email":"johndoe@nexus.org","firstName":"John","lastName":"Doe","pwd":"pipomolo","caps":{"USER_LIST":1,"VM_CREATE":1, "VM_SET_ACCESS":1}}' -H "Authorization: Bearer <AccessToken>"
func (r *RouterUsers) CreateUser(c echo.Context) error {
	// Retrieves logged user from context.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	if !user.HasCapability(caps.USERCAP_USER_CREATE) {
		return echo.NewHTTPError(http.StatusForbidden, msgInsufficientCaps)
	}

	// The client's parameters match exactly the User structure, thus
	// creates a new empty user to deserialize the client's parameters into.
	newUser := users.NewEmptyUser()

	// Deserializes the JSON body and checks its validity.
	if err := decodeJson(c, &newUser); err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	if err = newUser.Validate(); err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	newUser.Pwd = utils.HashPassword(newUser.Pwd)

	// Adds user and updates users file.
	if err = r.users.AddUser(newUser); err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	return c.NoContent(http.StatusCreated)
}

// Update a user's capabilities.
// Requires CAP_USER_SET_CAPS.
// curl --cacert ca.pem -X PUT http://localhost:1077/users/janedoes@nexus.org/caps -H 'Content-Type: application/json' -d '{"caps":{"USER_LIST":1,"VM_CREATE":1, "VM_SET_ACCESS":1}}' -H "Authorization: Bearer <AccessToken>"
func (r *RouterUsers) SetUserCaps(c echo.Context) error {
	// Retrieves logged user from context.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	if !user.HasCapability(caps.USERCAP_USER_SET_CAPS) {
		return echo.NewHTTPError(http.StatusForbidden, msgInsufficientCaps)
	}

	// Deserializes and validates client's parameters.
	p := new(params.UserSetCaps)
	if err := decodeJson(c, &p); err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	// Checks capabilities are valid.
	if err := p.Caps.Validate(); err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	// Retrieves the user for which the capabilities must be udpated.
	userToUpdate, err := r.users.GetUserByEmail(c.Param("email"))
	if err != nil {
		return echo.NewHTTPError(http.StatusNotFound, err.Error())
	}

	// Updates the user's capabilities.
	userToUpdate.Caps = p.Caps

	// Updates user and saves the new user file.
	if err = r.users.UpdateUser(&userToUpdate); err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	return c.NoContent(http.StatusOK)
}

// Update the logged user's password.
// curl --cacert ca.pem -X PUT http://localhost:1077/users/pwd -H 'Content-Type: application/json' -d '{"pwd": "new_password"}' -H "Authorization: Bearer <AccessToken>"
func (r *RouterUsers) SetUserPwd(c echo.Context) error {
	// Retrieves the logged user for which the password must be udpated.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	// Deserializes and validates client's parameters.
	p := new(params.UserSetPwd)
	if err := decodeJson(c, &p); err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	// Checks the new password is valid.
	if err := validator.New(validator.WithRequiredStructEnabled()).Struct(p); err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	// Hashes and udpates the user's password.
	user.Pwd = utils.HashPassword(p.Pwd)

	// Updates user and saves the new user file.
	if err = r.users.UpdateUser(user); err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	return c.NoContent(http.StatusOK)
}

// Unlock a user.
// Requires CAP_USER_UNLOCK.
// curl --cacert ca.pem -X PUT http://localhost:1077/users/janedoes@nexus.org/unlock -H 'Content-Type: application/json' -H "Authorization: Bearer <AccessToken>"
func (r *RouterUsers) UnlockUser(c echo.Context) error {
	// Retrieves logged user from context.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	if !user.HasCapability(caps.USERCAP_USER_UNLOCK) {
		return echo.NewHTTPError(http.StatusForbidden, msgInsufficientCaps)
	}

	userToUnlock, err := r.users.GetUserByEmail(c.Param("email"))
	if err != nil {
		return echo.NewHTTPError(http.StatusNotFound, err.Error())
	}

	if err = UnlockUser(c, userToUnlock.Email); err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	return c.NoContent(http.StatusOK)
}

// Reset a user's password.
// Requires CAP_USER_RESETPWD.
// curl --cacert ca.pem -X PUT http://localhost:1077/users/janedoes@nexus.org/resetpwd -H 'Content-Type: application/json' -H "Authorization: Bearer <AccessToken>"
func (r *RouterUsers) ResetPwd(c echo.Context) error {
	// Retrieves logged user from context.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	if !user.HasCapability(caps.USERCAP_USER_RESETPWD) {
		return echo.NewHTTPError(http.StatusForbidden, msgInsufficientCaps)
	}

	userToResetPwd, err := r.users.GetUserByEmail(c.Param("email"))
	if err != nil {
		return echo.NewHTTPError(http.StatusNotFound, err.Error())
	}

	pwdLength := 8
	pwdDigitCount := 4
	pwdSymbolCount := 0
	pwdRepeatChars := false
	randomPwd, err := utils.CustomPwd.Generate(pwdLength, pwdDigitCount, pwdSymbolCount, false, pwdRepeatChars)
	if err != nil {
		msg := "Failed to reset password for user " + userToResetPwd.Email + ": " + err.Error()
		log.Error(msg)
		return echo.NewHTTPError(http.StatusInternalServerError, err.Error())
	}

	userToResetPwd.Pwd = utils.HashPassword(randomPwd)

	// Updates user and saves the new user file.
	if err = r.users.UpdateUser(&userToResetPwd); err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	log.Info("User ", user.Email, " reset password for user ", userToResetPwd.Email)

	return c.JSON(http.StatusOK, params.UserSetPwd{Pwd: randomPwd})
}
