package router

import (
	"io"
	"net/http"

	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/server/exec"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/server/users"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/server/vms"

	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/common/caps"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/common/params"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/common/utils"

	"os"
	"path/filepath"

	"github.com/go-playground/validator/v10"
	"github.com/google/uuid"
	"github.com/labstack/echo/v4"
)

type (
	RouterVMs struct {
		users *users.Users
		vms   *vms.VMs
	}

	vmActionFn    func(c echo.Context, vm *vms.VM) error
	vmsListableFn func(c echo.Context, vm *vms.VM) error
)

func NewRouterVMs() *RouterVMs {
	return &RouterVMs{users: users.GetUsersInstance(), vms: vms.GetVMsInstance()}
}

// Returns all VMs for which any of these is true:
// - the user is the VM's owner
// - the user has at least one VMAccess capability in the VM
// - the user has at least one USERCAP_VM_xxx_ANY capability
// curl --cacert ca.pem -X GET https://localhost:1077/vms -H "Authorization: Bearer <AccessToken>"
func (r *RouterVMs) GetVMs(c echo.Context) error {
	// Retrieves logged user from context.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	/*
		// Convert a string containing VMAccess capabilities (separated by ",")
		// into an array of VMAccessCapability objects
		capsAsStringList := strings.Split(c.QueryParam("caps"), ",")
		capsList := make([]caps.VMAccessCapability, len(capsAsStringList))
		for i, capStr := range capsAsStringList {
			cap, success := caps.NewVMAccessCapability(capStr)
			if !success {
				return echo.NewHTTPError(http.StatusBadRequest, "invalid capability")
			}
			capsList[i] = cap
		}
	*/

	filterFunc := func(vm *vms.VM, userEmail string, userCaps caps.UserCapabilities) bool {
		if vm.IsOwner(userEmail) {
			return true
		} else if userCaps.HasVMxxxAnyCapability() {
			return true
		} else {
			vmAccessCaps, exists := vm.GetAccess()[userEmail]
			if exists {
				return len(vmAccessCaps) > 0
			} else {
				return false
			}
		}
	}

	return c.JSONPretty(http.StatusOK, r.vms.GetFilteredNetworkSerializedVMs(filterFunc, user.Email, user.Caps), "    ")
}

// Returns the specified VM.
// The VM is only returned if any of these is true:
// - the user is the VM's owner
// - the user has at least one VMAccess capability in the VM
// - the user has at least one USERCAP_VM_xxx_ANY capability
// curl --cacert ca.pem -X GET https://localhost:1077/vms/62ae8791-c108-4235-a7d6-074e9b6a9017 -H "Authorization: Bearer <AccessToken>"
func (r *RouterVMs) GetVM(c echo.Context) error {
	// Retrieves logged user from context.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	// Retrieves the VM based on its UUID.
	id, err := uuid.Parse(c.Param("id"))
	if err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}
	vm, err := r.vms.GetVM(id)
	if err != nil {
		return echo.NewHTTPError(http.StatusNotFound, err.Error())
	}

	if vm.IsOwner(user.Email) {
		return c.JSONPretty(http.StatusOK, vm.Serialize(), "    ")
	} else if user.Caps.HasVMxxxAnyCapability() {
		return c.JSONPretty(http.StatusOK, vm.Serialize(), "    ")
	} else {
		vmAccessCaps, exists := vm.GetAccess()[user.Email]
		if exists {
			if len(vmAccessCaps) > 0 {
				return c.JSONPretty(http.StatusOK, vm.Serialize(), "    ")
			} else {
				return echo.NewHTTPError(http.StatusForbidden, msgInsufficientCaps)
			}
		} else {
			return echo.NewHTTPError(http.StatusForbidden, msgInsufficientCaps)
		}
	}
}

// Returns attach credentials for VMs that are running and that can be attached to.
// Requires to be the VM's owner, or either capability:
// User cap: USERCAP_VM_ATTACH_ANY: returns VMs credentials for all running VMs.
// VM access cap: USERCAP_VM_ATTACH: returns VMs credentials for all running VMs with this cap for the logged user.
// curl --cacert ca.pem -X GET https://localhost:1077/vms/attach -H "Authorization: Bearer <AccessToken>"
func (r *RouterVMs) GetVMsAttachCredentials(c echo.Context) error {
	// Retrieves logged user from context.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	if user.HasCapability(caps.USERCAP_VM_ATTACH_ANY) {
		return c.JSONPretty(http.StatusOK, r.vms.GetVMAttachCredentialsSerialized(func(vm *vms.VM) bool {
			return vm.IsRunning()
		}), "    ")
	} else {
		// Returns all VMs:
		// - owned by the user
		// - for which the user has the specified capability (vmAccessCapability) in the VM's access
		return c.JSONPretty(http.StatusOK, r.vms.GetVMAttachCredentialsSerialized(func(vm *vms.VM) bool {
			if vm.IsOwner(user.Email) {
				return vm.IsRunning()
			} else {
				vmAccessCaps, exists := vm.GetAccess()[user.Email]
				if exists {
					_, found := vmAccessCaps[caps.VMCAP_ATTACH]
					return found && vm.IsRunning()
				} else {
					return false
				}
			}
		}), "    ")
	}
}

// Returns attach credentials for a VM specified by its UUID.
// Requires to be the VM's owner, or either capability:
// User cap: USERCAP_VM_ATTACH_ANY: returns VM credentials for the VM.
// VM access cap: USERCAP_VM_ATTACH: returns VM credentials for the VM with this cap for the logged user.
// curl --cacert ca.pem -X GET https://localhost:1077/vms/62ae8791-c108-4235-a7d6-074e9b6a9017/attach -H "Authorization: Bearer <AccessToken>"
func (r *RouterVMs) GetVMAttachCredentials(c echo.Context) error {
	// Retrieves logged user from context.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	// Retrieves the VM based on its UUID.
	id, err := uuid.Parse(c.Param("id"))
	if err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}
	vm, err := r.vms.GetVM(id)
	if err != nil {
		return echo.NewHTTPError(http.StatusNotFound, err.Error())
	}

	if !vm.IsRunning() {
		return echo.NewHTTPError(http.StatusBadRequest, "VM is not running")
	}

	// If user has USERCAP_VM_ATTACH_ANY capability, returns the VM.
	if user.HasCapability(caps.USERCAP_VM_ATTACH_ANY) {
		return c.JSONPretty(http.StatusOK, vm.SerializeAttachCredentials(), "    ")
	} else {
		if vm.IsOwner(user.Email) {
			return c.JSONPretty(http.StatusOK, vm.SerializeAttachCredentials(), "    ")
		} else {
			capabilities, exists := vm.GetAccess()[user.Email]
			if exists {
				_, found := capabilities[caps.VMCAP_ATTACH]
				if found {
					return c.JSONPretty(http.StatusOK, vm.SerializeAttachCredentials(), "    ")
				} else {
					return echo.NewHTTPError(http.StatusForbidden, msgInsufficientCaps)
				}
			} else {
				return echo.NewHTTPError(http.StatusForbidden, msgInsufficientCaps)
			}
		}
	}
}

// Returns the Spice credentials for the specific VM and specific VM attach password.
// Requires to be the VM's owner, or either capability:
// User cap: USERCAP_VM_ATTACH_ANY: returns VM credentials for the VM.
// VM access cap: USERCAP_VM_ATTACH: returns VM credentials for the VM with this cap for the logged user.
// curl --cacert ca.pem -X POST https://localhost:1077/vms/62ae8791-c108-4235-a7d6-074e9b6a9017/spicecreds -H 'Content-Type: application/json' -d '{"pwd":"46L8drgZ5Dx"}' -H "Authorization: Bearer <AccessToken>"
func (r *RouterVMs) VMSpiceCreds(c echo.Context) error {
	// Retrieves logged user from context.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	// Retrieves the VM based on its UUID.
	id, err := uuid.Parse(c.Param("id"))
	if err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}
	vm, err := r.vms.GetVM(id)
	if err != nil {
		return echo.NewHTTPError(http.StatusNotFound, err.Error())
	}

	if !vm.IsRunning() {
		return echo.NewHTTPError(http.StatusBadRequest, "VM is not running")
	}

	// Deserializes and validates client's parameters.
	p := new(params.VMAttachCreds)
	if err := decodeJson(c, &p); err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	// Does the received password match the "VM Attach" password stored in the VM?
	if p.Pwd != vm.Run.AttachPwd {
		return echo.NewHTTPError(http.StatusForbidden, "Access denied")
	}

	// If user has USERCAP_VM_ATTACH_ANY capability, returns the VM.
	if user.HasCapability(caps.USERCAP_VM_ATTACH_ANY) {
		return c.JSONPretty(http.StatusOK, vm.SerializeSpiceCredentials(), "    ")
	} else {
		if vm.IsOwner(user.Email) {
			return c.JSONPretty(http.StatusOK, vm.SerializeSpiceCredentials(), "    ")
		} else {
			capabilities, exists := vm.GetAccess()[user.Email]
			if exists {
				_, found := capabilities[caps.VMCAP_ATTACH]
				if found {
					return c.JSONPretty(http.StatusOK, vm.SerializeSpiceCredentials(), "    ")
				} else {
					return echo.NewHTTPError(http.StatusForbidden, msgInsufficientCaps)
				}
			} else {
				return echo.NewHTTPError(http.StatusForbidden, msgInsufficientCaps)
			}
		}
	}
}

// Returns the Spice credentials for the VM matching the specific VM attach password.
// Output: gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/common/vm.VMSpiceCredentialsSerialized
// Requires either to be the VM's owner, or either capability:
// USERCAP_EXAM_ATTACH or USERCAP_VM_ATTACH_ANY user capability:
// curl --cacert ca.pem -X POST https://localhost:1077/vms/spicecreds -H 'Content-Type: application/json' -d '{"pwd":"46L8drgZ5Dx"}' -H "Authorization: Bearer <AccessToken>"
func (r *RouterVMs) VMSpiceCredsAny(c echo.Context) error {
	// Retrieves logged user from context.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	if !(user.HasCapability(caps.USERCAP_VM_ATTACH_ANY) || user.HasCapability(caps.USERCAP_EXAM_ATTACH)) {
		return echo.NewHTTPError(http.StatusForbidden, "Access denied")
	}

	// Deserializes and validates client's parameters.
	p := new(params.VMAttachCreds)
	if err := decodeJson(c, &p); err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	vm, err := r.vms.GetVMByPwd(p.Pwd)
	if err != nil {
		return echo.NewHTTPError(http.StatusNotFound, err.Error())
	}

	return c.JSONPretty(http.StatusOK, vm.SerializeSpiceCredentials(), "    ")
}

// Creates a VM.
// Requires this capability:
// User cap: USERCAP_VM_CREATE.
// curl --cacert ca.pem -X POST https://localhost:1077/vms -H 'Content-Type: application/json' -d '{"name":"Systems Programming 2022","cpus":2,"ram":4096,"nic":"none","usbDevs":["1307:0165","1234:abcd"],"template":"Xubuntu_22.04"}' -H "Authorization: Bearer <AccessToken>"
func (r *RouterVMs) CreateVM(c echo.Context) error {
	// Retrieves logged user from context.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	if !user.HasCapability(caps.USERCAP_VM_CREATE) {
		return echo.NewHTTPError(http.StatusForbidden, msgInsufficientCaps)
	}

	// Deserializes and validates client's parameters.
	p := new(params.VMCreate)
	if err := decodeJson(c, &p); err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	// Creates a new VM from the client's parameters.
	vm, err := vms.NewVM(user.Email, p.Name, p.Cpus, p.Ram, p.Nic, p.UsbDevs, p.TemplateID, user.Email)
	if err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	if err = r.vms.AddVM(vm); err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	return c.JSONPretty(http.StatusCreated, vm.Serialize(), "    ")
}

// Starts a VM based on its ID.
// Requires to be the VM's owner, or either capability:
// User cap: USERCAP_VM_START_ANY: any VM can be started.
// VM access cap: USERCAP_VM_START: any of the VMs with this cap for the logged user can be started.
// curl --cacert ca.pem -X PUT https://localhost:1077/vms/e41f3556-ca24-4658-bd79-8c85bd6bff59/start -H "Authorization: Bearer <AccessToken>"
func (r *RouterVMs) StartVM(c echo.Context) error {
	return r.applyOnFilteredVMs(c, caps.USERCAP_VM_START_ANY, caps.VMCAP_START, func(c echo.Context, vm *vms.VM) error {
		// "" means a unique password to attach to the VM must be generated
		if err := r.vms.StartVM(vm.GetID(), ""); err != nil {
			return echo.NewHTTPError(http.StatusBadRequest, err.Error())
		}
		return c.NoContent(http.StatusOK)
	})
}

/*
// Unused, here to show how we would handle multiple actions in a single request.
func (r *RouterVMs) StartVMs(c echo.Context) error {
	return r.applyOnFilteredVMsMulti(c, caps.USERCAP_VM_START_ANY, caps.VMCAP_START, func(c echo.Context, vm *vms.VM) error {
		// "" means a unique password to attach to the VM must be generated
		if err := r.vms.StartVM(vm.GetID(), ""); err != nil {
			log.Debug("===> vm ", vm.GetID().String(), " starting failed!")
			return err
		}
		log.Debug("===> vm ", vm.GetID().String(), " started successfully")
		return nil
	})
}
*/

// Starts a VM based on its ID using the specified credentials (attach password).
// Requires to be the VM's owner, or either capability:
// User cap: USERCAP_VM_START_ANY: any VM can be started.
// VM access cap: USERCAP_VM_START: any of the VMs with this cap for the logged user can be started.
// curl --cacert ca.pem -X PUT https://localhost:1077/vms/e41f3556-ca24-4658-bd79-8c85bd6bff59/startwithcreds -H 'Content-Type: application/json' -d '{"pwd":"46L8drgZ5Dx"}' -H "Authorization: Bearer <AccessToken>"
func (r *RouterVMs) StartVMWithCreds(c echo.Context) error {
	return r.applyOnFilteredVMs(c, caps.USERCAP_VM_START_ANY, caps.VMCAP_START, func(c echo.Context, vm *vms.VM) error {
		// Deserializes and validates client's parameters.
		p := new(params.VMStartWithCreds)
		if err := decodeJson(c, &p); err != nil {
			return echo.NewHTTPError(http.StatusBadRequest, err.Error())
		}

		if err := r.vms.StartVM(vm.GetID(), p.Pwd); err != nil {
			return echo.NewHTTPError(http.StatusBadRequest, err.Error())
		}
		return c.NoContent(http.StatusOK)
	})
}

// Kills a VM based on its ID.
// Requires to be the VM's owner, or either capability:
// User cap: USERCAP_VM_STOP_ANY: any VM can be killed.
// VM access cap: USERCAP_VM_STOP: any of the VMs with this cap for the logged user can be killed.
// curl --cacert ca.pem -X PUT https://localhost:1077/vms/e41f3556-ca24-4658-bd79-8c85bd6bff59/stop -H "Authorization: Bearer <AccessToken>"
func (r *RouterVMs) KillVM(c echo.Context) error {
	return r.applyOnFilteredVMs(c, caps.USERCAP_VM_STOP_ANY, caps.VMCAP_STOP, func(c echo.Context, vm *vms.VM) error {
		if err := r.vms.KillVM(vm.GetID()); err != nil {
			return echo.NewHTTPError(http.StatusNotFound, err.Error())
		}
		return c.NoContent(http.StatusOK)
	})
}

// Gracefully shutdown a VM based on its ID.
// Requires to be the VM's owner, or either capability:
// User cap: USERCAP_VM_STOP_ANY: any VM can be shutdown.
// VM access cap: USERCAP_VM_STOP: any of the VMs with this cap for the logged user can be shutdown.
// curl --cacert ca.pem -X PUT https://localhost:1077/vms/e41f3556-ca24-4658-bd79-8c85bd6bff59/shutdown -H "Authorization: Bearer <AccessToken>"
func (r *RouterVMs) ShutdownVM(c echo.Context) error {
	return r.applyOnFilteredVMs(c, caps.USERCAP_VM_STOP_ANY, caps.VMCAP_STOP, func(c echo.Context, vm *vms.VM) error {
		if err := r.vms.ShutdownVM(vm.GetID()); err != nil {
			return echo.NewHTTPError(http.StatusNotFound, err.Error())
		}
		return c.NoContent(http.StatusOK)
	})
}

// Reboot a VM based on its ID.
// Requires to be the VM's owner, or either capability:
// User cap: USERCAP_VM_REBOOT_ANY: any VM can be rebooted.
// VM access cap: USERCAP_VM_REBOOT: any of the VMs with this cap for the logged user can be rebooted.
// curl --cacert ca.pem -X PUT https://localhost:1077/vms/e41f3556-ca24-4658-bd79-8c85bd6bff59/stop -H "Authorization: Bearer <AccessToken>"
func (r *RouterVMs) RebootVM(c echo.Context) error {
	return r.applyOnFilteredVMs(c, caps.USERCAP_VM_REBOOT_ANY, caps.VMCAP_REBOOT, func(c echo.Context, vm *vms.VM) error {
		if err := r.vms.RebootVM(vm.GetID()); err != nil {
			return echo.NewHTTPError(http.StatusNotFound, err.Error())
		}
		return c.NoContent(http.StatusOK)
	})
}

// Edit a VM' specs: name, cpus, ram, nic
// Requires to be the VM's owner, or either capability:
// User cap: USERCAP_VM_EDIT_ANY: any VM can be edited.
// VM access cap: USERCAP_VM_EDIT: any of the VMs with this cap for the logged user can be edited.
// curl --cacert ca.pem -X PUT https://localhost:1077/vms/e41f3556-ca24-4658-bd79-8c85bd6bff59 -H 'Content-Type: application/json' -d '{"name":"Edited VM","cpus":1,"ram":2048,"nic":"user","usbDevs":["1307:0165","1234:abcd"]}' -H "Authorization: Bearer <AccessToken>"
func (r *RouterVMs) EditVM(c echo.Context) error {
	return r.applyOnFilteredVMs(c, caps.USERCAP_VM_EDIT_ANY, caps.VMCAP_EDIT, func(c echo.Context, vm *vms.VM) error {
		// Deserializes and validates client's parameters.
		// Given these parameters are optional, we can't use a validator on them.
		// Validation is performed in vm.EditVM() instead.
		p := new(params.VMEdit)
		if err := decodeJson(c, &p); err != nil {
			return echo.NewHTTPError(http.StatusBadRequest, err.Error())
		}
		if err := r.vms.EditVM(vm.GetID(), p); err != nil {
			return echo.NewHTTPError(http.StatusBadRequest, err.Error())
		}
		return c.JSONPretty(http.StatusOK, vm.Serialize(), "    ")
	})
}

// Deletes a VM based on its ID.
// Requires to be the VM's owner, or either capability:
// User cap: USERCAP_VM_DESTROY_ANY: any VM can be deleted.
// VM access cap: USERCAP_VM_DESTROY: any of the VMs with this cap for the logged user can be deleted.
// curl --cacert ca.pem -X DELETE https://localhost:1077/vms/e41f3556-ca24-4658-bd79-8c85bd6bff59 -H "Authorization: Bearer <AccessToken>"
func (r *RouterVMs) DeleteVM(c echo.Context) error {
	return r.applyOnFilteredVMs(c, caps.USERCAP_VM_DESTROY_ANY, caps.VMCAP_DESTROY, func(c echo.Context, vm *vms.VM) error {
		if err := r.vms.DeleteVM(vm.GetID()); err != nil {
			return echo.NewHTTPError(http.StatusNotFound, err.Error())
		}
		return c.NoContent(http.StatusOK)
	})
}

// Set a VM access for a given user.
// Requires to be the VM's owner, or to have VM_SET_ACCESS_ANY capability, or to have BOTH capabilities:
// User cap: USERCAP_VM_SET_ACCESS
// VM access cap: USERCAP_VM_SET_ACCESS
// curl --cacert ca.pem -X PUT https://localhost:1077/vms/e41f3556-ca24-4658-bd79-8c85bd6bff59/caps/janedoes@nexus.org -H 'Content-Type: application/json' -d '{"caps":{"VM_LIST":1,"VM_START":1,"VM_STOP":1}}' -H "Authorization: Bearer <AccessToken>"
func (r *RouterVMs) SetVMAccessForUser(c echo.Context) error {
	// Retrieves logged user from context and checks she/he has sufficient capabilities.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	// Checks the user for which to set the VM Access actually exists.
	email := c.Param("email")
	_, err = r.users.GetUserByEmail(email)
	if err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	// Retrieves the vmID of the VM to modify.
	vmID, err := uuid.Parse(c.Param("id"))
	if err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	// Deserializes and validates client's parameters.
	p := new(params.VMAddAccess)
	if err := decodeJson(c, &p); err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}
	if err := validator.New(validator.WithRequiredStructEnabled()).Struct(p); err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	if err = r.vms.SetVMAccess(vmID, user, email, p.Access); err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	return c.NoContent(http.StatusOK)
}

// Delete a VM Access for a given user.
// Requires to be the VM's owner, or BOTH capabilities:
// User cap: USERCAP_VM_SET_ACCESS
// VM access cap: USERCAP_VM_SET_ACCESS
// curl --cacert ca.pem -X DELETE https://localhost:1077/vms/e41f3556-ca24-4658-bd79-8c85bd6bff59/caps/janedoes@nexus.org -H 'Content-Type: application/json' -H "Authorization: Bearer <AccessToken>"
func (r *RouterVMs) DeleteVMAccessForUser(c echo.Context) error {
	// Retrieves logged user from context and checks she/he has sufficient capabilities.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	// Retrieves the vmID of the VM to modify.
	vmID, err := uuid.Parse(c.Param("id"))
	if err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	// Purposedly does not check that the user to remove the VM access for actually
	// exists as it might have been deleted.
	email := c.Param("email")

	if err = r.vms.DeleteVMAccess(vmID, user, email); err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	return c.NoContent(http.StatusOK)
}

// Exports a VM's directory into a compressed archive.
// Requires to be the VM's owner, or either capability:
// User cap: VM_READFS_ANY: any VM can have their filesystem read.
// VM access cap: VM_READFS: any of the VMs with this cap for the logged user can have their filesystem read.
// curl --cacert ca.pem -X GET https://localhost:1077/vms/e41f3556-ca24-4658-bd79-8c85bd6bff59/exportdir -H 'Content-Type: application/json' -d '{"dir":"absolute_path_to_dir"}' -H "Authorization: Bearer <AccessToken>" --output dir.tar
func (r *RouterVMs) ExportVMDir(c echo.Context) error {
	return r.applyOnFilteredVMs(c, caps.USERCAP_VM_READFS_ANY, caps.VMCAP_READFS, func(c echo.Context, vm *vms.VM) error {
		vmID := vm.GetID().String()

		// Deserializes and validates the client's parameter.
		p := new(params.VMExportDir)
		if err := decodeJson(c, &p); err != nil {
			msg := "Failed extracting VM " + vmID + "'s dir: " + err.Error()
			log.Error(msg)
			return echo.NewHTTPError(http.StatusBadRequest, msg)
		}

		// Creates a unique tar filename.
		tarGzFile := filepath.Join(conf.Core.TmpDir, "exportdir_"+vm.GetID().String()+".tar.gz")

		// Extracts VM's p.Dir directory into tarGzFile on the host.
		if err := r.vms.ExportFilesFromVM(vm, p.Dir, tarGzFile); err != nil {
			msg := "Failed extracting VM " + vmID + "'s dir: " + err.Error()
			log.Error(msg)
			return echo.NewHTTPError(http.StatusBadRequest, msg)
		}

		// Sends the archive back to the client and once completed, deletes it.
		defer func(file string) {
			if err := os.Remove(file); err != nil {
				log.Error("Failed removing archive of extracted VM " + vmID + "'s dir: " + err.Error())
			}
		}(tarGzFile)
		return c.File(tarGzFile)
	})
}

// Import files into a VM's filesystem, in a specified directory. The file tree is received in a tar.gz archive.
// Requires to be the VM's owner, or either capability:
// User cap: VM_WRITEFS_ANY: any VM can import the file tree.
// VM access cap: VM_WRITEFS: any of the VMs with this cap for the logged user can import the file tree.
// curl --cacert ca.pem -X POST https://localhost:1077/vms/e41f3556-ca24-4658-bd79-8c85bd6bff59/importfiles -H 'Content-Type: multipart/form-data' -F vmDir="/home/nexus" -F file=@files.tar.gz -H "Authorization: Bearer <AccessToken>"
func (r *RouterVMs) ImportFilesToVM(c echo.Context) error {
	return r.applyOnFilteredVMs(c, caps.USERCAP_VM_WRITEFS_ANY, caps.VMCAP_WRITEFS, func(c echo.Context, vm *vms.VM) error {
		// Retrieves the various client arguments.
		vmDir := c.FormValue("vmDir")

		// Checks if importing files into the VM would be successfull.
		// The purpose of this function is to quickly checks if the import would succeed
		// without having to wait for the archive to be uploaded from the client (which can take a while).
		// It attempts to create an empty file inside the VM directory.

		// Create a local file on the host
		emptyLocalFile, err := utils.CreateRandomEmptyFile()
		if err != nil {
			return echo.NewHTTPError(http.StatusInternalServerError, err.Error())
		}

		// Remove the local file on the host when exiting the function
		defer os.Remove(emptyLocalFile)

		// Copy the local file from the host to the VM in vmDir
		if err := r.vms.ImportFilesToVM(vm, exec.CopyToVM, emptyLocalFile, vmDir); err != nil {
			return echo.NewHTTPError(http.StatusInternalServerError, err.Error())
		}

		// Remove the file we just copied in the VM
		filename := filepath.Base(emptyLocalFile)
		if err := r.vms.DeleteFileFromVM(vm, filepath.Join(vmDir, filename)); err != nil {
			return echo.NewHTTPError(http.StatusInternalServerError, err.Error())
		}

		// Retrieves the tar.gz archive (uploadedtarGzFile).
		tarGzFile, err := c.FormFile("file")
		if err != nil {
			return echo.NewHTTPError(http.StatusInternalServerError, err.Error())
		}

		src, err := tarGzFile.Open()
		if err != nil {
			return echo.NewHTTPError(http.StatusInternalServerError, err.Error())
		}
		defer src.Close()

		uuid, err := uuid.NewRandom()
		if err != nil {
			log.Error("Failed creating random UUID: " + err.Error())
			return echo.NewHTTPError(http.StatusInternalServerError, err.Error())
		}
		uploadedtarGzFile := filepath.Join(conf.Core.TmpDir, "upload_"+uuid.String()+".tar")
		dst, err := os.Create(uploadedtarGzFile)
		if err != nil {
			return echo.NewHTTPError(http.StatusInternalServerError, err.Error())
		}
		defer dst.Close()
		defer os.Remove(uploadedtarGzFile)

		if _, err = io.Copy(dst, src); err != nil {
			return echo.NewHTTPError(http.StatusInternalServerError, err.Error())
		}

		// Copy the archive's files into the VM
		// IMPORTANT: the VM cannot have its disk busy
		if err = r.vms.ImportFilesToVM(vm, exec.DearchiveToVM, uploadedtarGzFile, vmDir); err != nil {
			return echo.NewHTTPError(http.StatusBadRequest, err.Error())
		}

		return c.NoContent(http.StatusCreated)
	})
}

// Helper function that executes an action on a VM if either condition is true:
// - the logged user is the VM's owner (in this case, user has all VM access capabilities).
// - the logged user has the userCapabilityAny capability.
// - the VM access for the logged user matches the vmAccessCapability capability.
func (r *RouterVMs) applyOnFilteredVMs(c echo.Context, userCapabilityAny caps.UserCapability, vmAccessCapability caps.VMAccessCapability, action vmActionFn) error {
	// Retrieves the logged user from context.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	// Retrieves the VM on which to perform the action.
	id, err := uuid.Parse(c.Param("id"))
	if err != nil {
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}
	idStr := id.String()

	vm, err := r.vms.GetVM(id)
	if err != nil {
		return echo.NewHTTPError(http.StatusNotFound, "VM "+idStr+" not found: "+err.Error())
	}

	// If user is the VM's owner, executes the action.
	if vm.IsOwner(user.Email) {
		return action(c, vm)
		// If user has the XX_ANY capability, executes the action.
	} else if user.HasCapability(userCapabilityAny) {
		return action(c, vm)
		// Finally, if VM access for the logged user matches the required capability, executes the action.
	} else {
		userCaps, exists := vm.GetAccess()[user.Email]
		if !exists {
			return echo.NewHTTPError(http.StatusForbidden, "VM "+idStr+": "+msgInsufficientCaps)
		}
		_, hasAccess := userCaps[vmAccessCapability]
		if !hasAccess {
			return echo.NewHTTPError(http.StatusForbidden, "VM "+idStr+": "+msgInsufficientCaps)
		}
		return action(c, vm)
	}
}

// Unused, here to show how we would handle multiple actions in a single request.
/*
func (r *RouterVMs) applyOnFilteredVMsMulti(c echo.Context, userCapabilityAny caps.UserCapability, vmAccessCapability caps.VMAccessCapability, action vmActionFn) error {
	// Retrieves the logged user from context.
	user, err := getLoggedUser(r.users, c)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	query := c.QueryParam("id")
	if query == "" {
		log.Debug("==> missing id param")
		return echo.NewHTTPError(http.StatusBadRequest, "missing id query parameters")
	}

	var errorMsg strings.Builder
	error := false

	log.Debug("=> query=", query)
	vmIDs := strings.Split(query, ",")
	log.Debug("=> VM IDs length=", len(vmIDs))
	log.Debug("=> VM IDs=", vmIDs)

	for _, idStr := range vmIDs {
		log.Debug("==> vm id=", idStr)
		// Retrieves the VM on which to perform the action.
		id, err := uuid.Parse(idStr)
		if err != nil {
			log.Debug("==> parse error: ", "invalid ID: ", err.Error())
			errorMsg.WriteString("invalid ID: ")
			errorMsg.WriteString(err.Error())
			errorMsg.WriteString("\n")
			error = true
			continue
		}
		idStr := id.String()

		vm, err := r.vms.GetVM(id)
		if err != nil {
			errorMsg.WriteString("VM ")
			errorMsg.WriteString(idStr)
			errorMsg.WriteString(" not found: ")
			errorMsg.WriteString(err.Error())
			errorMsg.WriteString("\n")
			error = true
			continue
		}

		// If user is the VM's owner, executes the action.
		if vm.IsOwner(user.Email) {
			if err := action(c, vm); err != nil {
				errorMsg.WriteString(err.Error())
				errorMsg.WriteString("\n")
				error = true
				continue
			}
			// If user has the XX_ANY capability, executes the action.
		} else if user.HasCapability(userCapabilityAny) {
			if err := action(c, vm); err != nil {
				errorMsg.WriteString(err.Error())
				errorMsg.WriteString("\n")
				error = true
				continue
			}
			// Finally, if VM access for the logged user matches the required capability, executes the action.
		} else {
			userCaps, exists := vm.GetAccess()[user.Email]
			if !exists {
				errorMsg.WriteString("VM ")
				errorMsg.WriteString(idStr)
				errorMsg.WriteString(": ")
				errorMsg.WriteString(msgInsufficientCaps)
				errorMsg.WriteString("\n")
				error = true
				continue
			}
			_, hasAccess := userCaps[vmAccessCapability]
			if !hasAccess {
				errorMsg.WriteString("VM ")
				errorMsg.WriteString(idStr)
				errorMsg.WriteString(": ")
				errorMsg.WriteString(msgInsufficientCaps)
				errorMsg.WriteString("\n")
				error = true
				continue
			}
			if err := action(c, vm); err != nil {
				errorMsg.WriteString(err.Error())
				errorMsg.WriteString("\n")
				error = true
				continue
			}
		}
	}

	if error {
		return echo.NewHTTPError(http.StatusBadRequest, errorMsg.String())
	}

	return c.NoContent(http.StatusOK)
}
*/
